#
# © 2020 Mikhail Gusarov <dottedmag@dottedmag.net>
#
# This file is a part of twema and licensed under AGPLv3. See doc/COPYING at the
# root of the repository for the details.
#
import enum
import datetime
from collections import namedtuple
import json
import html


def quote_url(tweet, url):
    if not tweet["is_quote_status"]:
        return False
    return tweet["quoted_status_permalink"]["expanded"] == url


def whole_tweet_is_rt(tweet):
    return "retweeted_status" in tweet and tweet["full_text"].startswith("RT @")


ParsedUser = namedtuple("ParsedUser", ["name", "atname", "url"])


class ChunkKind(enum.Enum):
    TEXT = 0
    LINK = 1
    HIDDEN = 2


Chunk = namedtuple("Chunk", ["begin", "end", "kind", "text", "link"])


class Text:
    def __init__(self, text):
        self.text = text
        self.chunks = [Chunk(0, len(text), ChunkKind.TEXT, None, None)]

    def __str__(self):
        return "<twema.Text " + str(self.text) + " " + str(self.chunks) + ">"

    def link(self, begin, end, text, link):
        self._replace_chunk(
            begin, end, Chunk(begin, end, ChunkKind.LINK, text, link)
        )

    def hide(self, begin, end):
        self._replace_chunk(
            begin, end, Chunk(begin, end, ChunkKind.HIDDEN, None, None)
        )

    def hide_all(self):
        self.chunks = [Chunk(0, len(self.text), ChunkKind.HIDDEN, None, None)]

    def _replace_chunk(self, begin, end, new_chunk):
        i, chunk = self._find_chunk(begin, end)
        out = []
        if chunk.begin < begin:
            out.append(Chunk(chunk.begin, begin, ChunkKind.TEXT, None, None))
        out.append(new_chunk)
        if end < chunk.end:
            out.append(Chunk(end, chunk.end, ChunkKind.TEXT, None, None))
        self.chunks = self.chunks[:i] + out + self.chunks[i + 1 :]

    def _find_chunk(self, begin, end):
        for i, chunk in enumerate(self.chunks):
            if chunk.begin <= begin and begin < chunk.end:
                if chunk.begin < end and end <= chunk.end:
                    return i, chunk
                else:
                    raise RuntimeError(
                        "[{},{}) overlaps with chunk [{}, {})".format(
                            begin, end, chunk.begin, chunk.end
                        )
                    )
        raise RuntimeError("can't find chunk for [{}, {})".format(begin, end))

    def plaintext(self):
        out = ""
        for chunk in self.chunks:
            if chunk.kind == ChunkKind.TEXT:
                out += self.text[chunk.begin : chunk.end]
        return out

    def html(self):
        out = ""
        for chunk in self.chunks:
            if chunk.kind == ChunkKind.TEXT:
                out += self.text[chunk.begin : chunk.end]
            elif chunk.kind == ChunkKind.LINK:
                out += '<a href="' + chunk.link + '">' + chunk.text + "</a>"
        return out


class MediaType(enum.Enum):
    PHOTO = 0
    VIDEO = 1
    GIF = 2


def media_type(m):
    if (
        m["type"] == "video"
        or "video_info" in m
        or "additional_media_info" in m
    ):
        return MediaType.VIDEO
    if m["type"] == "animated_gif":
        return MediaType.GIF
    if m["type"] == "photo":
        return MediaType.PHOTO
    raise SyntaxError("Don't know how to handle " + s)


Media = namedtuple("Media", ["url", "type"])

ParsedTweet = namedtuple(
    "ParsedTweet",
    [
        "id",
        "url",
        "author",
        "text",
        "date",
        "quote",
        "retweet",
        "reply",
        "media",
    ],
)

Reply = namedtuple("Reply", ["id", "url"])


def parse_tweet(tweet):
    return _parse_tweet(json.loads(tweet))


def _parse_tweet(tweet):
    id = tweet["id_str"]
    author = ParsedUser(
        name=tweet["user"]["name"],
        atname="@" + tweet["user"]["screen_name"],
        url="https://twitter.com/" + tweet["user"]["screen_name"],
    )
    date = datetime.datetime.strptime(
        tweet["created_at"], "%a %b %d %H:%M:%S %z %Y"
    )
    text = Text(tweet["full_text"])

    entities = tweet.get("extended_entities")
    if entities is None:
        entities = tweet.get("entities")

    media = []
    for m in entities["media"] if entities and "media" in entities else []:
        mt = media_type(m)
        if mt == MediaType.VIDEO or mt == MediaType.GIF:
            media_url = m["expanded_url"]
        else:
            media_url = m["media_url_https"]
        media.append(Media(media_url, mt))

        assert len(m["indices"]) == 2
        medium_from = m["indices"][0]
        medium_to = m["indices"][1]
        text.hide(medium_from, medium_to)

    retweet = None
    if "retweeted_status" in tweet:
        retweet = _parse_tweet(tweet["retweeted_status"])
        text.hide_all()
        media = []

    quote, quote_url = None, None
    if retweet is None and tweet["is_quote_status"]:
        if "quoted_status" in tweet:
            quote = _parse_tweet(tweet["quoted_status"])
            quote_url = tweet["quoted_status_permalink"]["expanded"]
        else:
            # No quote, treat as a link
            pass

    for url in tweet.get("entities", {}).get("urls", {}):
        if url["expanded_url"] == quote_url:
            text.hide(url["indices"][0], url["indices"][1])
        else:
            text.link(
                url["indices"][0],
                url["indices"][1],
                url["display_url"],
                url["expanded_url"],
            )

    for mention in tweet.get("entities").get("user_mentions"):
        text.link(
            mention["indices"][0],
            mention["indices"][1],
            "@" + mention["screen_name"],
            "https://twitter.com/" + mention["screen_name"],
        )

    reply_id = tweet.get("in_reply_to_status_id_str")
    reply_author = tweet.get("in_reply_to_screen_name")

    reply = None
    if reply_id is not None:
        reply = Reply(
            reply_id,
            "https://twitter.com/" + reply_author + "/status/" + str(reply_id),
        )

    url = (
        "https://twitter.com/"
        + tweet["user"]["screen_name"]
        + "/status/"
        + str(id)
    )

    return ParsedTweet(
        id, url, author, text, date, quote, retweet, reply, media
    )


Thread = namedtuple("Thread", ["authors", "tweets"])


def collect_threads(tweets):
    threads = []
    tweets_to_threads = {}

    def find_thread(tweet):
        if tweet.id in tweets_to_threads:
            return tweets_to_threads[tweet.id]
        if tweet.retweet is not None and tweet.retweet.id in tweets_to_threads:
            return tweets_to_threads[tweet.retweet.id]
        if tweet.quote is not None and tweet.quote.id in tweets_to_threads:
            return tweets_to_threads[tweet.quote.id]
        if tweet.reply is not None and tweet.reply.id in tweets_to_threads:
            return tweets_to_threads[tweet.reply.id]

    for tweet in tweets:
        thread = find_thread(tweet)
        if thread is None:
            thread = []
            threads.append(thread)
        thread.append(tweet)
        tweets_to_threads[tweet.id] = thread
        if tweet.retweet is not None:
            tweets_to_threads[tweet.retweet.id] = thread
        if tweet.quote is not None:
            tweets_to_threads[tweet.quote.id] = thread
        if tweet.reply is not None:
            tweets_to_threads[tweet.reply.id] = thread

    out_threads = []
    for thread in threads:
        out_threads.append(
            Thread(
                list(set(map(lambda t: t.author, thread))),
                sorted(thread, key=lambda t: t.date),
            )
        )
    return out_threads


def have_tweet(tweet_id, thread):
    for thread_tweet in thread.tweets:
        if thread_tweet.id == tweet_id:
            return True
    return False


def tweet_ids(thread):
    ids = []
    for tweet in thread.tweets:
        ids.append(tweet.id)
    return ids


def parse_tweets(raw_tweets):
    tweets = []
    for raw_tweet in raw_tweets:
        tweets.append(parse_tweet(raw_tweet))
    return collect_threads(tweets)



#
# © 2020 Mikhail Gusarov <dottedmag@dottedmag.net>
#
# This file is a part of twema and licensed under AGPLv3. See doc/COPYING at the
# root of the repository for the details.
#
from collections import namedtuple
from email.message import EmailMessage

from jinja2 import Environment, select_autoescape, PackageLoader

from . import parse


def subjectify(s):
    return " ".join(s.split())


def format_subject(thread):
    subject = thread.authors[0].atname
    if len(thread.authors) > 1:
        subject += " (+" + str(len(thread.authors) - 1) + ")"

    t = subjectify(thread.tweets[0].text.plaintext())
    if t == "":
        if thread.tweets[0].retweet is not None:
            t = ": RT " + subjectify(thread.tweets[0].retweet.text.plaintext())
        else:
            subject += ": <no text>"
    else:
        subject += ": " + t

    if len(thread.tweets) > 1:
        subject += " (+" + str(len(thread.tweets) - 1) + ")"

    return subject


jenv = Environment(
    loader=PackageLoader("twema", "templates"),
    autoescape=select_autoescape(["html"]),
)

thread_template = jenv.get_template("thread.jinja")


def thread_html(config, thread):
    return thread_template.render(
        thread=thread,
        is_pic=lambda media: media.type == parse.MediaType.PHOTO,
        have_tweet=parse.have_tweet,
    )


def thread_email(config, thread):
    msg = EmailMessage()
    msg["Subject"] = format_subject(thread)
    msg["From"] = config["mail"]["from"]
    msg["To"] = config["mail"]["to"]

    msg.set_content("See HTML version")
    msg.add_alternative(thread_html(config, thread), subtype="html")

    return bytes(msg)



#
# © 2020 Mikhail Gusarov <dottedmag@dottedmag.net>
#
# This file is a part of twema and licensed under AGPLv3. See doc/COPYING at the
# root of the repository for the details.
#
import json

import twitter

from . import config


def connect(consumer_key, consumer_secret):
    f = config.prepare_oauth_tokens_file()
    if not config.have_oauth_tokens():
        twitter.oauth_dance("Twema", consumer_key, consumer_secret, f)
    oauth_token, oauth_token_secret = twitter.oauth.read_token_file(f)
    oauth = twitter.oauth.OAuth(
        oauth_token, oauth_token_secret, consumer_key, consumer_secret
    )
    return twitter.api.Twitter(auth=oauth)


def get_tweets(twitter, **args):
    return twitter.statuses.home_timeline(
        count=200, tweet_mode="extended", **args
    )


def get_tweets_after(twitter, id):
    tweets = {}
    ts = get_tweets(twitter, since_id=id)
    while len(ts):
        for t in ts:
            tweets[t["id"]] = json.dumps(t)
        smallest_id = min(ts, key=lambda s: s["id"])["id"]
        ts = get_tweets(twitter, since_id=id, max_id=smallest_id - 1)
    return tweets



#
# © 2020 Mikhail Gusarov <dottedmag@dottedmag.net>
#
# This file is a part of twema and licensed under AGPLv3. See doc/COPYING at the
# root of the repository for the details.
#
import sqlite3

from . import config

SCHEMA = """
CREATE TABLE IF NOT EXISTS tweets (
    id INTEGER PRIMARY KEY,
    content NOT NULL,
    sent NOT NULL DEFAULT 0
)
"""


def open():
    f = config.prepare_tweets_db_file()
    db = sqlite3.connect(f)
    db.execute(SCHEMA)
    db.commit()
    return db


def get_latest_saved_id(db):
    dt = db.execute("SELECT MAX(id) FROM tweets").fetchone()
    if dt:
        return dt[0]


def save_tweets(db, tweets):
    db.executemany(
        "INSERT INTO tweets (id, content) VALUES (?, ?)",
        [(k, v) for k, v in tweets.items()],
    )
    db.commit()


def get_tweet(db, id):
    row = db.execute("SELECT content FROM tweets WHERE id=?", [id]).fetchone()
    if row:
        return row[0]


def get_tweet_ids(db):
    for row in db.execute("SELECT id FROM tweets ORDER BY id").fetchall():
        yield row[0]


def get_unsent_tweets(db):
    for row in db.execute("SELECT content FROM tweets WHERE sent=0").fetchall():
        yield row[0]


def mark_tweets_as_sent(db, ids):
    db.executemany("UPDATE tweets SET sent=1 WHERE id=?", [(id,) for id in ids])
    db.commit()



#
# © 2020 Mikhail Gusarov <dottedmag@dottedmag.net>
#
# This file is a part of twema and licensed under AGPLv3. See doc/COPYING at the
# root of the repository for the details.
#
import subprocess
from email.utils import parseaddr


def send(config, msg):
    [name, email] = parseaddr(config["mail"]["from"])

    p = subprocess.Popen(
        ["/usr/sbin/sendmail", "-t", "-oi", "-f", email, "-F", name],
        stdin=subprocess.PIPE,
    )
    p.communicate(msg)
    return p.returncode == 0



#
# © 2020 Mikhail Gusarov <dottedmag@dottedmag.net>
#
# This file is a part of twema and licensed under AGPLv3. See doc/COPYING at the
# root of the repository for the details.
#
import sys
import argparse

from . import config
import twema


def main():
    parser = argparse.ArgumentParser(description="Twitter to e-mail gateway")
    parser.add_argument(
        "command", metavar="(fetch|send)", help="the command to run"
    )
    parser.add_argument(
        "--verbose", help="enable verbose output", action="store_true"
    )
    parser.add_argument(
        "--id", help="comma-separated IDs of tweet to render or 'all'"
    )
    args = parser.parse_args()

    c = config.load()

    if args.command == "fetch":
        twema.fetch(c)
    elif args.command == "send":
        twema.send(c)
    elif args.command == "render-email":  # debugging commands
        twema.render_email(c, args.id)
    elif args.command == "render-html":
        twema.render_html(c, args.id)
    elif args.command == "print-raw":
        twema.print_raw(c, args.id)
    elif args.command == "print-parsed-tweet":
        twema.print_parsed_tweet(c, args.id)
    elif args.command == "list":
        twema.list(c)
    else:
        print("Unknown command:", args.command)
        sys.exit(1)



#
# © 2020 Mikhail Gusarov <dottedmag@dottedmag.net>
#
# This file is a part of twema and licensed under AGPLv3. See doc/COPYING at the
# root of the repository for the details.
#
import pathlib
import sys
import os.path

import toml
import xdg

config_file_path = xdg.XDG_CONFIG_HOME / "twema" / "config.toml"
oauth_tokens_path = xdg.XDG_DATA_HOME / "twema" / "oauth-tokens"
tweets_db_path = xdg.XDG_DATA_HOME / "twema" / "tweets.db"


def _load():
    try:
        return toml.load(config_file_path)
    except FileNotFoundError as e:
        return {}


def load():
    data = _load()
    if not isinstance(data.get("mail"), dict) or (
        "from" not in data["mail"] or "to" not in data["mail"]
    ):
        print("mail.from and mail.to are required in " + str(config_file_path))
        sys.exit(1)

    if not isinstance(data.get("app"), dict) or (
        "consumer_key" not in data["app"]
        or "consumer_secret" not in data["app"]
    ):
        print(
            "app.consumer_key and app.consumer_secret are required in "
            + str(config_file_path)
        )
        sys.exit(1)
    return data


def have_oauth_tokens():
    return os.path.exists(oauth_tokens_path)


def prepare_oauth_tokens_file():
    oauth_tokens_path.parent.mkdir(parents=True, exist_ok=True)
    return oauth_tokens_path


def prepare_tweets_db_file():
    tweets_db_path.parent.mkdir(parents=True, exist_ok=True)
    return tweets_db_path



#
# © 2020 Mikhail Gusarov <dottedmag@dottedmag.net>
#
# This file is a part of twema and licensed under AGPLv3. See doc/COPYING at the
# root of the repository for the details.
#
from setuptools import setup

setup(
    name="twema",
    version="0.5.0",
    description="Twitter-to-email gateway",
    url="https://github.com/dottedmag/twema",
    author="Mikhail Gusarov",
    author_email="dottedmag@dottedmag.net",
    license="AGPLv3",
    packages=["twema"],
    entry_points={"console_scripts": ["twema=twema.cli:main"],},
    include_package_data=True,
    install_requires=[
        "Jinja2==2.11.0",
        "toml==0.10.0",
        "twitter==1.18.0",
        "xdg==4.0.1",
    ],
)



#
# © 2020 Mikhail Gusarov <dottedmag@dottedmag.net>
#
# This file is a part of twema and licensed under AGPLv3. See doc/COPYING at the
# root of the repository for the details.
#
import sys

from . import db, formatting, twitter, parse, mail


def fetch(config):
    d = db.open()
    t = twitter.connect(
        config["app"]["consumer_key"], config["app"]["consumer_secret"],
    )

    latest_saved_id = db.get_latest_saved_id(d)
    if latest_saved_id:
        tweets = twitter.get_tweets_after(t, latest_saved_id)
    else:
        tweets = twitter.get_tweets(t)

    db.save_tweets(d, tweets)


def send(config):
    d = db.open()

    raw_tweets = db.get_unsent_tweets(d)

    threads = parse.parse_tweets(raw_tweets)

    for thread in threads:
        msg = formatting.thread_email(config, thread)
        if mail.send(config, msg):
            db.mark_tweets_as_sent(d, parse.tweet_ids(thread))


# debugging commands


def get_raw_tweets(config, ids):
    d = db.open()

    if ids is None:
        ids = []
    elif ids == "all":
        ids = db.get_tweet_ids(d)
    else:
        ids = ids.split(",")

    return [db.get_tweet(d, id) for id in ids]


def render_email(config, ids):
    raw_tweets = get_raw_tweets(config, ids)

    threads = parse.parse_tweets(raw_tweets)

    for thread in threads:
        sys.stdout.buffer.write(formatting.thread_email(config, thread))


def render_html(config, ids):
    raw_tweets = get_raw_tweets(config, ids)

    threads = parse.parse_tweets(raw_tweets)

    for thread in threads:
        sys.stdout.write(formatting.thread_html(config, thread))


def list(config):
    d = db.open()
    for id in db.get_tweet_ids(d):
        print(id)


def print_raw(config, id):
    d = db.open()
    print(db.get_tweet(d, id))


def print_parsed_tweet(config, id):
    d = db.open()
    tweet = parse.parse_tweet(db.get_tweet(d, id))
    print(tweet)
    print(tweet.text)
